/* Newsfeed Syndication program:HShivanan 
 * Started: Aug 24 2002
 *
 * 
 * For licensing terms, see the file COPYING
 * For contact information, see the file AUTHORS
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "net.h"
#include "stuff.h"
#include "xmlparser.h"
#include "configparser.h"
#include "fn.h"
#include "report.h"
#include "stream.h"
#include "connection.h"
#include "mail.h"

#define OUTPUT_NORMAL    1
#define OUTPUT_VERBOSE   2
#define OUTPUT_QUIET     3


static CHANNEL_LIST *current_channel;
static int output_type=OUTPUT_NORMAL;
static int show_errors=1;
static FILE *out_file;

/*Add a node to the CHANNEL_LIST linked list*/
void add_channel(char *_url,char *_description,char *_type)
{
	char *url,*description;
	size_t url_len,desc_len;
	CHANNEL_LIST *channel=(CHANNEL_LIST *)fn_malloc(sizeof(CHANNEL_LIST));
	ASSERT(_url!=NULL);
	url_len=strlen(_url);
	url=(char *)fn_malloc(url_len+1);
	memcpy(url,_url,url_len+1);
	if (_description)
	{
		desc_len=strlen(_description);
		description=(char *)fn_malloc(desc_len+1);
		memcpy(description,_description,desc_len+1);
	}
	else
	{
		description=NULL;
	}
	ASSERT(url!=NULL);
	if (current_channel->url==NULL &&
			current_channel->next==NULL)
	{
		current_channel->url=url;
		current_channel->description=description;
		current_channel->type=NULL; /*not yet implemented*/
	}else
	{
	channel->url=url;
	channel->description=description;
	channel->type=NULL;
	current_channel->next=channel;
	channel->next=NULL;
	current_channel=channel;
	}
}



void verbose(const char *text,...)
{
	va_list args;
	if (output_type!=OUTPUT_VERBOSE)return;
	va_start(args, text);
	vfprintf(out_file,text,args);
	va_end(args);
}
void print(const char *text,...)
{
	va_list args;
	if (output_type==OUTPUT_QUIET)return;
	va_start(args, text);
	vfprintf(out_file,text,args);
	va_end(args);
}
void error(const char *text,...)
{
	va_list args;
	va_start(args,text);
	if (show_errors)
		vfprintf(stderr,text,args);
	va_end(args);
}

/* *channel must be preallocated*/
void generate_output(ITEM_LIST *il,FN_OPTIONS *fn)
{
	ASSERT(fn->template_file!=NULL);
 	verbose("Generating report from template:%s\n",fn->template_file);
	generate_report(il,fn);
 	return;
}

/*Skip the HTTP header information*/
int find_offset(char *buffer,size_t len)
{
	int i=0;
	while(i<len-4)
	{
		if (buffer[i]=='\r' && buffer[i+1]=='\n' &&
				buffer[i+2]=='\r' && buffer[i+3]=='\n'
				)return i+4;
		i++;
	}
	return -1;
}

/*Download and parse a news feed*/
int fetch_news(ITEM_LIST **il,char *channel,char *description,
		FN_OPTIONS *fn)
{
	int i;
	char *buffer;
	STREAM *stream;
	if (!strncmp(channel,"http://",7))
	{
		stream = fetch_url(fn,channel);
		if (!stream || !(stream->buffer))
		{
			return -1;
		}
		buffer = stream->buffer;
		if (buffer==NULL)
		{
			error("%s doesn't seem to have any data\nThis is wierd\n",channel);
			return -1;
		}
		i = find_offset(buffer,strlen(buffer));
		verbose(".\n");
		if (i<=0)
		{
			error("Unable to parse HTTP header:%i\nThis is really wierd\n",i);
			fn_free(buffer);
			return -1;
		}
	}else
	{
		verbose("Reading local file:%s...\n",channel);
		buffer=read_file(channel);
		i=0;
		if (buffer==NULL)
		{
			error("Unable to open local file:%s\n",channel);
			return -1;
		}
	}
	
	if (description==NULL)
		parse_newsfeed(il,buffer+i,channel,NULL);
	else
		parse_newsfeed(il,buffer+i,channel,description);
	
	if (has_option(OPT_SAVE_XML))
	{
		save_file(buffer+i,"raw",strlen(buffer)-i);
		print("saving in raw...\n");
	}
	
	
	/* TODO: Fix memory leak here */
	if (has_option(OPT_SAVE_ALL))
	{
		char * path;
		char * clean_file;
		size_t f_len;
		clean_file = convert_url_to_file (channel);
		f_len = strlen (clean_file);
		
		if (fn->xml_dir)
		{
			size_t p_len = strlen (fn->xml_dir);
			path = (char *)fn_malloc (p_len+f_len+1);
			memcpy (path, fn->xml_dir, p_len);
			memcpy (path+p_len, clean_file, f_len);
			path[p_len+f_len]='\0';
		}else
		{
			path = clean_file;
		}
		save_file (buffer+i, path, strlen(buffer)-i);
		print ("saving in %s...\n", path);
	}

	
	fn_free(buffer);
	return 1;
}

void free_channel_list(CHANNEL_LIST *channels)
{
	CHANNEL_LIST *tmp,*c=channels;
	while (c!=NULL)
	{
		tmp=c->next;
		if (c->url!=NULL)
			fn_free(c->url);
		if (c->description)
			fn_free(c->description);
		if (c->type)
			fn_free(c->type);
		fn_free(c);
		c=tmp;
	}
}

#define _FREE(x) if (fn->x) fn_free(fn->x)
void free_fn_options(FN_OPTIONS *fn)
{
	
	free_channel_list (fn->channels);
	free_score_list (fn->score_list);
	free_http_request (fn->http_request);
	free_mail_info (fn->mail_info);
	
	_FREE (proxy);
	_FREE (channel_file);
	_FREE (template_file);
	_FREE (config_file);
	_FREE (address);
	_FREE (score_file);
	_FREE (report_file);
	_FREE (raw_file);
	_FREE (xml_dir);
	fn_free(fn);
}

FN_OPTIONS *new_options()
{
	FN_OPTIONS *fn=fn_malloc(sizeof(FN_OPTIONS));

	/*Zero out all members
	 * TODO: Is this portable?
	 */
	memset(fn,'\0',sizeof(FN_OPTIONS));
	fn->options=0;
	fn->score_list=new_score_list(10);
	fn->http_request = new_http_request();
	fn->mail_info = new_mail_info();
	return fn;
}

/*Perform messy proxy verification*/

#define PROXY_ERROR_MSG "You have not specified a proxy correctly.\nEx: -proxy \"proxy.server:8080\"\n"
#define ENV_ERROR_MSG   "The http_proxy environment variable seems to be incorrect.\n"\
	"Use --no-proxy to override it or explicitly specify one.\n"
static int verify_proxy(FN_OPTIONS *fn, int from_env)
{
	if (fn->proxy)
	{
		char **p = NULL;
		int offset = 0;
		
		verbose("Proxy found.Validating syntax...\n");
		
		if (!strncmp(fn->proxy,"http://",7))
			offset = 7;
		
		p=split_str((fn->proxy)+offset,":",2);
		
		if ((p[0]==NULL)||(p[1]==NULL))
		{
			error(from_env?ENV_ERROR_MSG:PROXY_ERROR_MSG);
			if (p[0])
				fn_free(p[0]);
			if (p[1])
				fn_free(p[1]);
			free(p);
			return -1;
		}
		
		fn->proxy=p[0];
		
		if ((fn->proxy_port=atoi(p[1]))==0)
		{
			error(from_env?ENV_ERROR_MSG:PROXY_ERROR_MSG);
			fn_free(p[1]);
			fn_free(p);
			return -1;
		}
		fn_free(p[1]);
		fn_free(p);
	}

	return 1;
}

/*Used when we need to flush out the current
  channel list and start fresh*/
void refresh_channel_list(FN_OPTIONS *fn)
{
	free_channel_list(fn->channels);
	fn->channels=fn_malloc(sizeof(CHANNEL_LIST));
	current_channel=fn->channels;
	current_channel->url=NULL;
	current_channel->description=NULL;
	current_channel->type=NULL;
	current_channel->next=NULL;
}

/* Horrendous macros to aid in command-line parsing*/
#define parse_arg_val(opt,struct_val); if (strcmp(*argv,opt)==0){\
				if (!(*++argv)){\
         		fprintf(stderr,"An argument should proceed the \"%s\" option\n",\
		    	*(argv-1));\
			    return -1;}\
				fn->struct_val = duplicate_str ((*argv));\
				argv++;\
				break;}
											
#define parse_arg_op(opt,code); if (strcmp(*argv,opt)==0){\
				fn->options|=(code);\
				argv++;\
				break;}
				
#define undefined() fprintf(stderr,"Unrecognized option:%s\n",*argv);\
					return -1

int parse_cmdline(int argc,char **argv,FN_OPTIONS *fn)
{
	/*Get passed the program name*/
	argv++;
	while ((*argv)!=NULL)
	{
		if ((*argv)[0]=='-')
		{
			switch ((*argv)[1])
			{
#ifdef DEBUG
				case 'a':parse_arg_op("-a",OPT_SHOW_ALLOC_STATS);undefined();
#endif
				case 'p':parse_arg_val("-proxy",proxy);
						 undefined();
				case 't':parse_arg_val("-template",template_file);
						 undefined();
				case 'o':parse_arg_val("-o",report_file);
						 parse_arg_val("-output",report_file);
						 undefined();
				case 'h':parse_arg_op("-h",OPT_HELP);undefined();
				case 'i':if (strcmp("-i",*argv)==0)
							 {
								 verbose("Ignoring channels from file.\n");
								 refresh_channel_list(fn);
								 argv++;
								 break;
							 }
						 undefined();
				case 'v':parse_arg_op("-v",OPT_VERBOSE);
						 undefined();
				case 'V':parse_arg_op("-V",OPT_VERSION);undefined();
				case 's':parse_arg_op("-s",OPT_SAVE_ALL);undefined();
				case 'd':parse_arg_op("-d",OPT_DUMP_REPORT);undefined();
				case 'q':parse_arg_op("-q",OPT_QUIET);undefined();
				case 'Q':parse_arg_op("-Q",OPT_QUIET|OPT_SUPPRESS_ERRORS);undefined();
				case 'r':parse_arg_op("-r",OPT_SAVE_XML);undefined();
				case '-':
						 switch((*argv)[2])
						 {
							 case 'd':parse_arg_op("--dump-raw",OPT_SAVE_XML);
									  parse_arg_op("--dump-report",OPT_DUMP_REPORT);
									  undefined();

							 case 'h':parse_arg_op("--help",OPT_HELP);
									  undefined();
									  
							 case 'i':if (strcmp("--ignore",*argv)==0)
							 		  {
								 			verbose("Ignoring channels from file.\n");
								 			refresh_channel_list(fn);
								 			argv++;
								 			break;
							 		  }
						 			  undefined();
									  
							 case 'n':parse_arg_op("--no-sort",OPT_NO_SORT);
									  parse_arg_op("--no-proxy",OPT_NO_PROXY);
									  undefined();
									  
							 case 'q':parse_arg_op("--quiet",OPT_QUIET);
									  undefined();
									  
							 case 'r':parse_arg_op("--raw",OPT_SAVE_XML);
									  undefined();
									  
							 case 's':parse_arg_op("--save-xml",OPT_SAVE_ALL);
#ifdef DEBUG
								  parse_arg_op("--show-alloc-stats",OPT_SHOW_ALLOC_STATS);
#endif
									  undefined();
									  
							 case 'v':parse_arg_op("--version",OPT_VERSION);
								  parse_arg_op("--verbose",OPT_VERBOSE);
								  parse_arg_op("--very-quiet",OPT_QUIET|OPT_SUPPRESS_ERRORS);  
									  undefined();
									  
							 case 'x':parse_arg_val ("--xml-dir",xml_dir);
									  undefined();
									  
							 default:undefined();
						 }
						 break;
				default:undefined();
			}
		}
		else
		{
			add_channel(*argv,NULL,NULL);
			argv++;
		}
	}
	
	/*if the user has requested version or help info, don't
	  bother verifying anything*/
	if (has_option(OPT_VERSION)||has_option(OPT_HELP))
		return 1;
    
	/*Check to see if we're running in verbose
	  or quiet mode*/
	if (has_option(OPT_VERBOSE))
		output_type = OUTPUT_VERBOSE;
	if (has_option(OPT_QUIET))
		output_type = OUTPUT_QUIET;
	if (has_option(OPT_SUPPRESS_ERRORS))
		show_errors = 0;
	
	
	/*If we need to send the report to stdout, send everything else
	 * to stderr */
	if (has_option(OPT_DUMP_REPORT))
		out_file=stderr;
		
	
	
	return 1;
}


/* 
 * Sanity check various parameters to make sure they are 
 * legitimate 
 * 
 */
int sanity_check (FN_OPTIONS * fn)
{
	/*perform verification on various files*/
	if (fn->template_file!=NULL)
	{
		if (!file_exists(fn->template_file))
		{
			error("Unable to find file:%s\n",fn->template_file);
			return -1;
		}
	}
	else
	{
		error("You need to specified a template.\n");
		error("Use -template <file> or -h to see all possible commands\n");
		return -1;
	}
	/* TODO: verify other files */
	if (fn->xml_dir)
	{
		/* TODO: verify existence of directory */

		/* Make sure we have a trailing '/' */
		char * b;
		size_t len = strlen (fn->xml_dir);
		if (fn->xml_dir[len-1] != FOLDER_CHAR)
		{
			b = (char *)fn_malloc (len+2);
			memcpy (b,fn->xml_dir,len);
			b[len+1] = FOLDER_CHAR;
			b[len+2] = '\0';
			fn->xml_dir = b;
		}
	}
	
	/* validate the proxy server*/
	if (!has_option(OPT_NO_PROXY))
	{
		int from_env = 0;
		/* if no proxy was specified, try and get one from the
		 * http_proxy environment variable */
		if (!(fn->proxy))
		{
			fn->proxy = getenv(ENV_HTTP_PROXY);
			from_env = 1;
		}

		if (verify_proxy(fn, from_env)==-1)
			return -1;
	}
	return 1;
}

void show_options()
{
	printf("   Options are as follows:\n");
	printf("  -h,--help             Displays this message\n");
	printf("  -d,--dump-report      Dump report to stdout\n");
	printf("  -r,--dump-raw         Dump xml data to 'raw'\n");
	printf("  -s,--save-xml         Save xml data to individual files\n");
	printf("  -v,--verbose          Run in verbose mode\n");
	printf("  -q,--quiet            Run in quiet mode\n");
	printf("  -Q,--very-quiet       Quiet mode + suppression of error reporting\n");
	printf("  -i,--ignore-rc        Only take channels from the command line\n");
	printf("  --no-sort             Don't sort items based on score\n");
	printf("  -template <file>      Use <file> as the template file\n");
	printf("  -o,-output <file>     The report file to be generated(overrides -d)\n");
	printf("  -proxy <server:port>  Use HTTP proxy server <server> at port <port>\n");
	printf("  --no-proxy            Don't use any proxy server\n");
/* TODO
 * printf("  -t,--timeout <count>  Timeout (in seconds) when retrieving news feeds\n");
 */
	printf("  -V,--version          Show version information\n");
#ifdef DEBUG
	printf("  -a,--show-alloc-stats Show memory allocation stats (debug release only)\n");
#endif
}
void show_usage()
{
	printf("Usage:\n%s [options] [newsfeeds...]\n",PROG_NAME);
	printf("Version:%s\nBuild Date:%s\n",VERSION,BUILD_DATE);
#ifdef DEBUG
	printf("Debug Build\n");
#endif
}
char * get_rc_path ()
{
	STREAM * s;
	char *home = NULL;
	char *result = NULL;

	/*First check the current directory*/
	if (file_exists(FNRC))
		return FNRC;
	
	home = getenv(ENV_HOME);
	if (!home)
		return NULL;

	s = stream_create (10);
	stream_write (s, home, strlen(home));
	stream_write (s, "/", 1);
	stream_write (s, FNRC, FNRC_LEN);
	stream_write (s, "\0", 1);
	if (!file_exists (s->buffer))
	{
		stream_free (s);
		return NULL;
	}
	result = fn_malloc (s->size);
	memcpy (result,s->buffer,s->size);
	stream_free (s);
	return result;
	
}
int main(int argc,char **argv)
{
	ITEM_LIST * il;
	CHANNEL_LIST * ch;
	char * rcfile;
	int count;
	
	/*For debug build*/
	int show_alloc_stats=0;
	
	FN_OPTIONS *fn = new_options();
	
	fn->channels = (CHANNEL_LIST *)fn_malloc(sizeof(CHANNEL_LIST));
	current_channel = fn->channels;
	il = new_item_list(10);
	il->count = 0;
	out_file = stdout;

	rcfile = get_rc_path ();
	if (rcfile)
		parse_config_file (rcfile ,fn);
	
	/*If we don't have a valid command-line, exit*/
	if (parse_cmdline (argc,argv,fn) < 0)
		return -1;
	
	if (has_option(OPT_HELP)||has_option(OPT_VERSION))
	{
		show_usage();
		if has_option (OPT_HELP)
			show_options();
		goto clean_up;
		free_item_list(il);
		free_fn_options(fn);
		goto done;
	}

	if (sanity_check (fn) < 0)
		return -1;

#ifdef DEBUG
	if (has_option(OPT_SHOW_ALLOC_STATS))
		show_alloc_stats=1;
#endif

	ch = fn->channels;

	
	if (ch==NULL || ch->url == NULL)
	{
		error("No channels were specified\n");
		goto clean_up;
	}

	
	
	/*Check if we need to dump anything to stdout.
	  If so, redirect 'other' outputs to stderr*/
	/*Walk the linked-list, downloading the feed*/
	
	
	count = 0;
	
	while ((ch!=NULL)&&(ch->url!=NULL))
	{
		print("Fetching news from %s\n",ch->url);
		if (fetch_news(&il,ch->url,ch->description,fn)>0)
		{
			verbose("Done\n");
			count++;
		}
		ch = ch->next;
	}

	if (count <=0 )
	{
		print("No newsfeeds were downloaded\n");
		goto clean_up;
	}

	verbose ("%i feeds were retrieved\n",count);

	
	/*TODO: Make sure we have something in il*/
	
	/* fn->template_file should be guaranteed to be filled
	 * at this point
	 */
	//template = read_file(fn->template_file);
	if (!has_option(OPT_NO_SORT))
	{
		verbose("Scoring and sorting news items...\n");
		score_items(il,fn->score_list);
	}
	generate_output(il,fn);
	
	/*Clean up*/
clean_up:
	free_item_list(il);
	free_fn_options(fn);
done:
	if (show_alloc_stats)
		print_stats();
	return 1;
}
